#include "StdAfx.h"
#include "SObserver.h"

using namespace pml;

static const wchar_t g_procFilenameAndParams[] =  L"\\SessionMonitor.exe\" -app ";

CSObserver::CSObserver()
{
}

CSObserver::~CSObserver()
{
    for(ProcessHandlesMap::iterator hProcPoolIt = m_hProcessesPool.begin(); hProcPoolIt != m_hProcessesPool.end(); ++hProcPoolIt)
    {
        TerminateProcess(hProcPoolIt->second, 0);
        CloseHandle(hProcPoolIt->second);
    }
}

void CSObserver::GotLogon(const std::wstring& userName, int sid)
{
    boost::mutex::scoped_lock lock(m_mutex);
    boost::shared_ptr<CThreadForSession> threadForSess(new CThreadForSession(sid, this));
    ThreadPtr threadPtr (new boost::thread(boost::bind(&CThreadForSession::Execute, threadForSess)) );
    ThreadHolder threadHolder(threadForSess, threadPtr);
    std::pair<ThreadsMap::iterator, bool> resultThread = m_threadsPool.insert(std::make_pair(sid, threadHolder));
    if(!resultThread.second)
    {
        throw std::runtime_error("Cannot insert thread to the pool");
    }
    resultThread.first->second.SetStopEvent(true);
}

void CSObserver::GotLogoff(int sid)
{
    boost::mutex::scoped_lock lock(m_mutex);
    m_threadsPool.erase(sid);

    ProcessHandlesMap::iterator hProcPoolIt = m_hProcessesPool.find(sid); 
    if(hProcPoolIt != m_hProcessesPool.end())
    {
        if(!TerminateProcess(hProcPoolIt->second, 0))
        {
            // TODO: add log entry ("Cannot terminate process err = 0x%X", GetLastError());
        }
        else
        {
            CloseHandle(hProcPoolIt->second);
            m_hProcessesPool.erase(sid);
        }
    }
}

void CSObserver::PutProcessHandleIntoMap(CHandleGuard &hProcGuard, int sid)
{
    boost::mutex::scoped_lock lock(m_mutex);
    std::pair<ProcessHandlesMap::iterator, bool> result = m_hProcessesPool.insert(std::make_pair(sid, hProcGuard.get()));
    if(!result.second)   
    {
        throw std::runtime_error("Cannot insert process handle to the pool");
    }
    hProcGuard.release();
}

std::wstring pml::CreateCommandLineStr()
{
    wchar_t fileNameBuf[1024] = {0};
    if(!GetModuleFileNameW(0, fileNameBuf, sizeof(fileNameBuf) / sizeof(wchar_t) - 1))
    {
        throw std::runtime_error("CreateRecItProcessInSessionId: GetModuleFileNameW failed");
    }

    std::wstring fileName(fileNameBuf);
    std::wstringstream commandLine;
    commandLine << L"\"" << strUtil::GetFilePath(fileName) << g_procFilenameAndParams;
    return commandLine.str();
}

int pml::CreateProcessInSessionId (int sid, CHandleGuard &hProcessGuard, CHandleGuard &hTokenGuard, const std::wstring & commandLine)
{
    HANDLE hDpToken(0);
    if(!DuplicateTokenEx(hTokenGuard.get(), TOKEN_ALL_ACCESS, NULL, SecurityImpersonation, TokenPrimary, &hDpToken))
    {
        throw std::runtime_error("CreateRecItProcessInSessionId: DuplicateTokenEx failed");
    }
    CHandleGuard hDpTokenGuard(hDpToken);

    if(!SetTokenInformation(hDpTokenGuard.get(), TokenSessionId, &sid, sizeof(sid)))
    {
        throw std::runtime_error("CreateRecItProcessInSessionId: SetTokenInformation failed");
    }

    STARTUPINFOW si;
    memset( &si, 0, sizeof(si));
    si.cb = sizeof(si);
    si.lpDesktop = L"Winsta0\\default";
    PROCESS_INFORMATION processInfo;
    memset( &processInfo, 0, sizeof(processInfo) );

    CAccessAttributes aa;
    aa.InitAttributes();

    if(!CreateProcessAsUser(hDpTokenGuard.get(), NULL, const_cast<wchar_t *>(commandLine.c_str()), &aa.GetAttributes(), &aa.GetAttributes(), false, CREATE_NO_WINDOW, NULL, NULL, &si, &processInfo))
    {
        return GetLastError();
    }
    SetKernelObjectSecurity(processInfo.hProcess, DACL_SECURITY_INFORMATION, aa.GetSecurityDescriptor());

    CloseHandle(processInfo.hThread);
    hProcessGuard.reset(processInfo.hProcess);
    return 0;
}

bool pml::GetProcessToken(int sid, CHandleGuard &hTokenGuard)
{
    DWORD pids[1024*10], cbNeeded, cProcesses;

    if ( !EnumProcesses( pids, sizeof(pids), &cbNeeded ) )
        return false;

    // Calculate how many process identifiers were returned.
    cProcesses = cbNeeded / sizeof(DWORD);
    for(DWORD i = 0; i<cProcesses; ++i)
    {
        DWORD dwPid = pids[i];
        HANDLE hProcess = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwPid);
        if (hProcess)
        {
            CHandleGuard guard(hProcess);
            HANDLE hToken = 0;
            if (OpenProcessToken(hProcess, TOKEN_ALL_ACCESS, &hToken))
            {
                CHandleGuard token(hToken);

                try
                {
                    std::wstring name = GetTokenUser(hToken);
                    DWORD dwSession = GetTokenSessionId(hToken);

                    const wchar_t arg[] = L"NT AUTHORITY\\";
                    if (_wcsnicmp(name.c_str(),arg, sizeof(arg)/sizeof(arg[0])-1)==0)
                    {
                        if(sid == dwSession)
                        {
                            hTokenGuard.reset(token.release());
                            return true;
                        }
                    }
                }
                catch(const std::exception &ex)
                {
                }
            }
        }
    }
    return false;
}

extern "C"
BOOL WINAPI ConvertSidToStringSidW(PSID Sid, LPWSTR* StringSid);

DWORD pml::GetTokenSessionId(HANDLE TokenHandle)
{
    DWORD id = 0, length;
    BOOL res = GetTokenInformation(TokenHandle,
        TokenSessionId,
        &id,
        sizeof(id),
        &length);
    if (res)
        return id;
    throw std::runtime_error("Invalid token");
}

std::wstring pml::GetTokenUser(HANDLE Token)
{
    DWORD tmp;

    std::wstring userName;

    SID_NAME_USE snUse;
    DWORD sidNameSize = 64;
    std::vector<WCHAR> sidName;
    sidName.resize(sidNameSize);

    DWORD sidDomainSize = 64;
    std::vector<WCHAR> sidDomain;
    sidDomain.resize(sidNameSize);

    DWORD userTokenSize = 1024;
    std::vector<WCHAR> tokenUserBuf;
    tokenUserBuf.resize(userTokenSize);
    TOKEN_USER *userToken = (TOKEN_USER *)&tokenUserBuf.front();

    if( GetTokenInformation( Token, TokenUser, userToken, userTokenSize, &tmp ) )
    {
        if( LookupAccountSidW( NULL, userToken->User.Sid, &sidName.front(), &--sidNameSize, &sidDomain.front(), &--sidDomainSize, &snUse ) )
        {
            userName = &sidDomain.front();
            userName += L"\\";
            userName += &sidName.front();
        }
        else
        {
            WCHAR *pSidString;
            if(ConvertSidToStringSidW(userToken->User.Sid, &pSidString))
            {
                CLocalAllocGuard memGuard(pSidString);
                userName = pSidString;
            }
            else
                throw std::runtime_error(__FUNCTION__"() - ConvertSidToStringSidW fail");
        }
    }
    else
        throw std::runtime_error(__FUNCTION__"() - GetTokenInformation fail");

    return userName;
}
